"""Task metrics."""

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, Protocol

import django
from django.dispatch import receiver
from opentelemetry import metrics

if TYPE_CHECKING:
    from opentelemetry.util.types import Attributes

if django.VERSION >= (6, 0):
    from django.tasks import signals as django_native_signals  # type: ignore[import-untyped]
else:
    django_native_signals = None

try:
    from django_tasks import signals as django_tasks_signals
except ImportError:
    django_tasks_signals = None  # type: ignore[assignment]


if django_native_signals or django_tasks_signals:
    meter: metrics.Meter = metrics.get_meter(__name__)

    tasks_enqueued: metrics.Counter = meter.create_counter(
        "django_tasks_enqueued",
        description="Number of enqueued tasks",
    )
    tasks_started: metrics.Counter = meter.create_counter(
        "django_tasks_started",
        description="Number of started tasks",
    )
    tasks_success: metrics.Counter = meter.create_counter(
        "django_tasks_success",
        description="Number of tasks finished with success",
    )
    tasks_failed: metrics.Counter = meter.create_counter(
        "django_tasks_failed",
        description="Number of tasks finished failing",
    )
    tasks_pending: metrics.UpDownCounter = meter.create_up_down_counter(
        "django_tasks_pending",
        description="Number of currently pending tasks",
    )
    tasks_running: metrics.UpDownCounter = meter.create_up_down_counter(
        "django_tasks_running",
        description="Number of currently running tasks",
    )

    class _Task(Protocol):
        queue_name: str

    class _TaskResult(Protocol):
        status: Literal["READY", "RUNNING", "FAILED", "SUCCEEDED"]
        task: _Task

    class _TaskBackend(Protocol):
        alias: str

    def on_task_enqueued(*, sender: type[_TaskBackend], task_result: _TaskResult, **_kwargs: Mapping[str, Any]) -> None:
        """Increment the number of enqueued and pending tasks."""
        task_attributes: Attributes = {
            "backend": sender.alias,
            "queue_name": task_result.task.queue_name,
        }
        tasks_enqueued.add(1, task_attributes)
        tasks_pending.add(1, task_attributes)

    def on_task_started(*, sender: type[_TaskBackend], task_result: _TaskResult, **_kwargs: Mapping[str, Any]) -> None:
        """Increment the number of started and running tasks."""
        task_attributes: Attributes = {
            "backend": sender.alias,
            "queue_name": task_result.task.queue_name,
        }
        tasks_started.add(1, task_attributes)
        tasks_pending.add(-1, task_attributes)
        tasks_running.add(1, task_attributes)

    def on_task_finished(*, sender: type[_TaskBackend], task_result: _TaskResult, **_kwargs: Mapping[str, Any]) -> None:
        """Increment the number of enqueued and pending tasks."""
        task_attributes: Attributes = {
            "backend": sender.alias,
            "queue_name": task_result.task.queue_name,
        }

        metric = tasks_failed if task_result.status == "FAILED" else tasks_success
        metric.add(1, task_attributes)
        tasks_running.add(-1, task_attributes)

    if django_native_signals:
        receiver(django_native_signals.task_enqueued)(on_task_enqueued)
        receiver(django_native_signals.task_started)(on_task_started)
        receiver(django_native_signals.task_finished)(on_task_finished)

    if django_tasks_signals:
        receiver(django_tasks_signals.task_enqueued)(on_task_enqueued)
        receiver(django_tasks_signals.task_started)(on_task_started)
        receiver(django_tasks_signals.task_finished)(on_task_finished)



"""Read by Django to configure :mod:`django_opentelemetry`."""

from django.apps import AppConfig
from django.db.models import options
from django.utils.translation import gettext_lazy as _

from . import features

if features.should_use_model_meta() and "metric_label" not in options.DEFAULT_NAMES:
    options.DEFAULT_NAMES = (*options.DEFAULT_NAMES, "metric_label")


class OpenTelemetryConfig(AppConfig):
    """:mod:`django_opentelemetry` app configuration."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_opentelemetry"
    verbose_name = _("OpenTelemetry")

    def ready(self) -> None:
        """Register signals."""
        if features.should_collect_model_metrics():
            from . import model_signals  # noqa: F401, PLC0415

        if features.should_collect_task_metrics():
            from . import task_signals  # noqa: F401, PLC0415

        return super().ready()



from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import InMemoryMetricReader


def create_meter_provider() -> tuple[MeterProvider, InMemoryMetricReader]:
    """Create a `MeterProvider` and an `InMemoryMetricReader`.

    Returns:
        A tuple with the meter provider in the first element and the
        in-memory metrics exporter in the second
    """
    memory_reader = InMemoryMetricReader()
    meter_provider = MeterProvider(metric_readers=[memory_reader])
    return meter_provider, memory_reader



import pytest
from opentelemetry import metrics
from opentelemetry.sdk.metrics.export import MetricsData

from .models import SimpleModel
from .utils import create_meter_provider

pytestmark = pytest.mark.django_db


def test_model_creation() -> None:
    meter_provider, memory_metrics_reader = create_meter_provider()
    metrics.set_meter_provider(meter_provider)

    SimpleModel.objects.create(name="Bob")

    metric_data: MetricsData | None = memory_metrics_reader.get_metrics_data()

    assert metric_data is not None
    assert isinstance(metric_data, MetricsData)
    assert len(metric_data.resource_metrics) > 0

    for resource_metric in metric_data.resource_metrics:
        for scope_metric in resource_metric.scope_metrics:
            for metric in scope_metric.metrics:
                if metric.name == "django_model_inserts":
                    data_points = list(metric.data.data_points)
                    assert len(data_points) == 1
                    assert data_points[0].attributes["model"] == "simple_model"  # type: ignore[attr-defined]
                    return

                if metric.name in ["django_model_updates", "django_model_deletes"]:
                    pytest.fail(f"Should not get {metric.name} metric")

    pytest.fail("Metric not found")


def test_model_update() -> None:
    instance = SimpleModel.objects.create(name="Bob")

    meter_provider, memory_metrics_reader = create_meter_provider()
    metrics.set_meter_provider(meter_provider)

    instance.name = "Mark"
    instance.save()

    metric_data: MetricsData | None = memory_metrics_reader.get_metrics_data()

    assert metric_data is not None
    assert len(metric_data.resource_metrics) > 0

    for resource_metric in metric_data.resource_metrics:
        for scope_metric in resource_metric.scope_metrics:
            for metric in scope_metric.metrics:
                if metric.name == "django_model_updates":
                    data_points = list(metric.data.data_points)
                    assert len(data_points) == 1
                    assert data_points[0].attributes["model"] == "simple_model"  # type: ignore[attr-defined]
                    return

                if metric.name in ["django_model_inserts", "django_model_deletes"]:
                    pytest.fail(f"Should not get {metric.name} metric")

    pytest.fail("Metric not found")


def test_model_delete() -> None:
    instance = SimpleModel.objects.create(name="Bob")

    meter_provider, memory_metrics_reader = create_meter_provider()
    metrics.set_meter_provider(meter_provider)

    instance.delete()

    metric_data: MetricsData | None = memory_metrics_reader.get_metrics_data()

    assert metric_data is not None
    assert len(metric_data.resource_metrics) > 0

    for resource_metric in metric_data.resource_metrics:
        for scope_metric in resource_metric.scope_metrics:
            for metric in scope_metric.metrics:
                if metric.name == "django_model_deletes":
                    data_points = list(metric.data.data_points)
                    assert len(data_points) == 1
                    assert data_points[0].attributes["model"] == "simple_model"  # type: ignore[attr-defined]
                    return

                if metric.name in ["django_model_inserts", "django_model_updates"]:
                    pytest.fail(f"Should not get {metric.name} metric")

    pytest.fail("Metric not found")



"""Read settings."""

from django.conf import settings


def should_use_model_meta() -> bool:
    """Return if model meta should be used to setup metrics.

    Defaults to True.
    """
    return getattr(settings, "OTEL_MODEL_META", True)


def should_collect_model_metrics() -> bool:
    """Return if model metrics should be collected.

    Defaults to True.
    """
    return getattr(settings, "OTEL_MODEL_METRICS", True)


def extra_models() -> dict[str, str]:
    """Return any configured extra models, with it's label."""
    return getattr(settings, "OTEL_EXTRA_MODELS", {})


def should_collect_task_metrics() -> bool:
    """Return if task metrics should be collected.

    Defaults to True.
    """
    return getattr(settings, "OTEL_TASK_METRICS", True)



"""Model metrics."""

from collections.abc import Mapping
from typing import Any

from django.db import models
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from opentelemetry import metrics

from . import features

meter: metrics.Meter = metrics.get_meter(__name__)

model_inserts: metrics.Counter = meter.create_counter(
    "django_model_inserts",
    description="Number of insert operations by model.",
)
model_updates: metrics.Counter = meter.create_counter(
    "django_model_updates",
    description="Number of update operations by model.",
)
model_deletes: metrics.Counter = meter.create_counter(
    "django_model_deletes",
    description="Number of delete operations by model.",
)


@receiver(post_save)
def on_model_insert_or_update(
    *,
    sender: type[models.Model],
    created: bool,
    **_kwargs: Mapping[str, Any],
) -> None:
    """Increment the either the insert or update metric for a given model."""
    if metric_label := getattr(sender._meta, "metric_label", None) or features.extra_models().get(sender._meta.label):
        metric: metrics.Counter = model_inserts if created else model_updates
        metric.add(1, {"model": metric_label})


@receiver(post_delete)
def on_model_delete(
    *,
    sender: type[models.Model],
    **_kwargs: Mapping[str, Any],
) -> None:
    """Increment the delete metric for a given model."""
    if metric_label := getattr(sender._meta, "metric_label", None) or features.extra_models().get(sender._meta.label):
        model_deletes.add(1, {"model": metric_label})



from django.db import models


class SimpleModel(models.Model):
    name = models.CharField(max_length=100)

    class Meta:
        metric_label: str = "simple_model"

    def __str__(self) -> str:
        return str(self.pk)
